from pydantic import BaseModel
from typing import Optional

from sqlalchemy import Column, Integer, String

from gear_score import coefficients
from gear_score.enums import RarityEnum, SlotEnum
from gear_score.db import Base as SqliteBase


class Item(BaseModel):
    """Simple representation of an item"""

    level: int
    id: int
    name: str
    rarity: RarityEnum
    slot: SlotEnum
    gs: Optional[int]

    def __init__(self, **kwargs):
        """Custom init is needed as pydantic doesn't adequately work
        with native python getters & setters"""
        super().__init__(**kwargs)

        gs = kwargs.get("gs") or self._count_gs()
        setattr(self, "gs", gs)

    def _count_gs(self):
        gear_score = (
            (self._level - self._gs_scale[self._rarity]["subtrahend"])
            / self._gs_scale[self._rarity]["divisor"]
            * self._quality_coefficient
            * self._slot_coefficient
        )
        if gear_score < 0:
            gear_score = 0
        return gear_score

    @property
    def _quality_coefficient(self) -> float:
        """Quality coefficient differs depending on item rarity"""
        _scale_modifier = 1
        _constant_scale = 1.8618
        if self.rarity == RarityEnum.legendary:
            _scale_modifier = 1.3
        elif self.rarity in (RarityEnum.poor, RarityEnum.common):
            _scale_modifier = 0.005

        return _scale_modifier * _constant_scale

    @property
    def _slot_coefficient(self) -> float:
        """Different coefficient for each slot"""
        return coefficients.SLOT_COEFFICIENTS[self.slot]

    @property
    def _rarity(self) -> RarityEnum:
        """GS formulae uses mutated item rarity"""
        _formulae_rarity_mapping = {
            RarityEnum.poor: RarityEnum.uncommon,
            RarityEnum.common: RarityEnum.uncommon,
            RarityEnum.legendary: RarityEnum.epic,
            RarityEnum.heirloom: RarityEnum.rare,
        }
        return _formulae_rarity_mapping.get(self.rarity, self.rarity)

    @property
    def _level(self) -> float:
        """All heirloom items have fixed item level"""
        _level = self.level
        if self.rarity == RarityEnum.heirloom:
            _level = 187.05
        return _level

    @property
    def _gs_scale(self) -> dict:
        _gs_scale = coefficients.GS_COEFFICIENTS["LOW_LEVEL"]
        if self._level > 120:
            _gs_scale = coefficients.GS_COEFFICIENTS["HIGH_LEVEL"]

        return _gs_scale

    class Config:
        orm_mode = True


class FailedItem(BaseModel):
    """Returned by getters for items that were not found by
    PublicDbGetter and LocalDbGetter if raise_exceptions=False"""
    id: int

    @property
    def gs(self) -> float:
        return 0.0


class DbItem(SqliteBase):
    __tablename__ = "item"
    """SqlAlchemy representation of an item"""
    id = Column(Integer, primary_key=True)
    name = Column(String)
    level = Column(Integer)
    rarity = Column(Integer)
    slot = Column(Integer)
    gs = Column(Integer)

"""
CREATE TABLE `item` (
	`id` INT NOT NULL,
	`level` INT NOT NULL,
	`rarity` INT NOT NULL,
	`slot` INT NOT NULL,
	`gs` INT NOT NULL,
	`name` CHAR NOT NULL,
	PRIMARY KEY (`id`)
);
"""


from .enums import RarityEnum, SlotEnum

# Resulting GS of item of certain type is multiplied by the corresponding coefficient below
SLOT_COEFFICIENTS = {
    SlotEnum.head: 1,
    SlotEnum.neck: 0.5625,
    SlotEnum.shoulder: 0.75,
    SlotEnum.body: 0,
    SlotEnum.chest: 1,
    SlotEnum.waist: 0.75,
    SlotEnum.legs: 1,
    SlotEnum.feet: 0.75,
    SlotEnum.wrist: 0.5625,
    SlotEnum.hands: 0.75,
    SlotEnum.fingers: 0.5625,
    SlotEnum.trinkets: 0.5625,
    SlotEnum.weapon: 1,
    SlotEnum.shield: 1,
    SlotEnum.gun: 0.3164,
    SlotEnum.cloaks: 0.5625,
    SlotEnum.two_handed: 2,
    SlotEnum.tabard: 0,
    SlotEnum.main_hand: 1,
    SlotEnum.off_hand: 1,
    SlotEnum.ammunition: 0,
    SlotEnum.relics: 0.3164
}

# The data struct is used in GS calculation. See gear_score.models.Item._count_gs
GS_COEFFICIENTS = {
    'HIGH_LEVEL': {
        RarityEnum.epic: {'subtrahend': 91.45, 'divisor': 0.65},
        RarityEnum.rare: {'subtrahend': 81.375, 'divisor': 0.8125},
        RarityEnum.uncommon: {'subtrahend': 73, 'divisor': 1}
    },
    'LOW_LEVEL': {
        RarityEnum.epic: {'subtrahend': 26, 'divisor': 1.2},
        RarityEnum.rare: {'subtrahend': 0.75, 'divisor': 1.8},
        RarityEnum.uncommon: {'subtrahend': 8, 'divisor': 2},
        RarityEnum.common: {'subtrahend': 0, 'divisor': 2.25}
    }
}



import setuptools

readme = open("README.md").read()

setuptools.setup(
    name="gear_score",
    version="0.0.3",
    author="Eugene Prodan",
    author_email="mora9715@gmail.com",
    description="Python implementation of GearScore WoW addon",
    long_description=readme,
    long_description_content_type='text/markdown',
    url="https://github.com/mora9715/gearscore",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers"

    ],
    python_requires='>=3.6',
    install_requires=[
        'aiohttp',
        'pydantic',
        'sqlalchemy',
        'aiosqlite'
    ]
)


from pydantic import BaseSettings

import gear_score


class Settings(BaseSettings):
    use_local_db: bool = True
    public_db_hosts: list = ("https://wotlk.ezhead.org", "https://wotlkdb.com")
    db_dsn = f'sqlite+aiosqlite:///{gear_score.__name__}.sqlite'
    http_request_timeout = 5

    class Config:
        env_prefix = 'gs'


settings = Settings()



from typing import List

from gear_score.models import Item
from gear_score.getters import LocalDbGetter, PublicDbGetter


async def get_items_from_local_or_remote(*args: int) -> List[Item]:
    """
    :param args: list of items ids to be queried
    :return:
    """
    items = []
    failed_items = []

    # ROUND 1: look in local DB
    local_getter = LocalDbGetter()
    local_results = await local_getter.get(*args, many=True, raise_exceptions=False)
    for result in local_results:
        if isinstance(result, Item):
            items.append(result)
        else:
            failed_items.append(result)

    if failed_items:
        remote_getter = PublicDbGetter()
        remote_results = await remote_getter.get(
            *[i.id for i in failed_items],
            many=True,
            raise_exceptions=False,
            save_to_db=True
        )
        items += remote_results

    return items


def count_gs(*args: Item) -> int:
    _gs = 0
    for item in args:
        _gs += int(item.gs)

    return _gs
